;
; OSMC uMOB Control Software and PCB Design files
;
; Copyright 2001, 2002 Larry Barello
;
; Author: Larry Barello
;         larry@barello.net
;
; Revision History
;	05/12/2002 - Larry Barello
;	- Flip needs to both swap left/right *and* reverse motor direction.  The 
;	  latter never was done (apparently this function was never used...)
;	04/12/2002 - Larry Barello
;	- In the init code, I set left_hold, right_hold and aux_hold to RC_RANGE/2.
;	  this was fine for the previous version where the holding registers held
;	  only the raw data.  THIS version, however, the holding registers hold
;	  signed values, so RC_RANGE/2 was equal to full tilt forward.  Oops.
;	- Changed order of resetting RC pulse logic to prevent slight race condition
;	  This had no apparent effect on the operation of the code.
;	03/11/2002 - Larry Barello
;	- The limit routine needed to be ported to 16 bit math to handle the larger
;	  dynamic range with the expanded resolution.  +/- 200 (in mixed mode) was
;	  too big for byte math.
;	02/26/2002 - Larry Barello
;	- split apart deadband from limit code so input signal is deadbanded and computed
;	  (PWM) signal is limited.  Prior the deadband was applied to the computed signal
;	  and this created artifacts in mixed mode near zero.
;	- Corrected support for the LMD18200 H-Bridge.  It looks like each H-Bridge chip has
;	  a slightly different topology so I updated the conditional code to facilitate future
;	  modification.
;	02/25/2002 - Larry Barello
;	- Modified timer reload code in interrupt handler to dynamically compensate for
;	  variable interrupt latency.  Noticable improvement in stability of R/C to PWM (less dither)
;	- Unified OSMC and LMD18200 versions of code through compile time options in makefile
;	- actually read the spec for the LMD18200 to get the signals correct.
;	02/22/2002 - Larry Barello
;	- Modified so PWM is not inverted in reverse to accomodate
;	  users wanting to drive off the shelf power amplifiers (e.g. LM18200)
;	01/08/2002 - Larry Barello
;	- adjusted missing pulse limit from 10 to 7 to accomodate
;	  weird UK radio with a 45hz update rate (david gilson)
;	12/03/2001 - Larry Barello
;	- Shipped version 1.1
;	11/21/2001 - Larry Barello
;	- Added copyright notice
;	11/04/2001 - Larry Barello
;	- Moved "wdr" instruction to before the enabling of WDT.
;	- Moved update of holding register to branch where pulse is known
;	  to be valid.  Still doesn't detect "missing" pulse, but at least
;	  it doesn't stuff bogus values into the holding register.
;	- Initalized the holding registers to mid-stick on startup.
;	9/30/2001 - Larry Barello.  Shipped version 1.0
;       8/28/2001 - Larry Barello.  Initial version of Micro MOB controller
;
; GNU avr-gcc 3.0
;
; Define the hardware interface and bits
;
#include "io1200.h"

#define BV(A) (1<<(A))
/*
The resolution of our input is determined by the RC_RANGE value.
RC_RANGE indicates the number of slices in 1.0 ms, which is the
usable range of an typical R/C pulse.  With RC_RANGE = 200, our
resolution is 5 us.  The clock is driven at the CPU rate (1x)
The timer reload value is calculated in the interrupt routine since
there is a variable interrupt latency depending upon what instruction
was being executed when the timer interrupts.  (at 200 khz only 10-12
instructions are exectued between interrupts so I can't guarentee
being interrupted out of the "sleep" instruction.

The useful range is set by PWM_RANGE, which matches the resolution of
the input function.

Watchdog is enabled.   and is reset after receiving VALID_PULSE number
of valid pulses (e.g. 1.0 - 2.0 ms range)

 */
#define CPUCLK		12000000
#define RC_RANGE 	200			// +/- 100 or 5 us/slice
#define PWM_RANGE	64			// +/- 64
#define TICKRATE	(RC_RANGE * 1000)	// rc_range/1.0ms
#define TCCR0_INIT	BV(CS00)               	// CPUCLK/1
#define TCNT0_INIT	(256-(CPUCLK/TICKRATE)+9)
#define	TIMSK_INIT	(1<<TOIE0)
#define MCUCR_INIT	(1<<SE)      		// Enable sleep mode
#define WDTCR_INIT 	(BV(WDP2) | BV(WDE))	// ~.25 second timeout
;
; PWM output is +/- 64 for a PWM rate of ~3khz and R/C pulse range of 1000us
;
#define RC_OFFSET       (-RC_RANGE)   		// 1.0ms/tickrate = 100 * 10us = 1ms
#define AUX_LOW_LIM	(RC_RANGE * 33 / 100)	// 1/3 ~ stick half back
#define AUX_HIGH_LIM	(RC_RANGE * 66 / 100)	// 2/3 ~ stick half forward
#define DEADBAND_LOW	(20 * RC_RANGE/1000)	// 20 us
#define DEADBAND_HIGH	(60 * RC_RANGE/1000)	// 60 us
#define VALID_PULSE	7			// Number of consecutive valid pulses needed

;
; Port pin definitions
;
; Crazy assignments are for easy PCB layout...
;
#define RC_RIGHT  PINB,  PB0	// IN  - Enable Pull-up
#define DIS_RIGHT PORTB, PB1	// OUT - High
#define AUX_HI    PORTB, PB2	// OUT - low
#define AUX_LO    PORTB, PB3	// OUT - low
#define FLIP	  PINB,  PB4	// IN  - Enable Pull-up
#define DEADBAND  PINB,  PB5	// IN  - Enable pull-up
#define BRAKE     PINB,  PB6	// IN  - Enable pull-up
#define MIXED     PINB,  PB7	// IN  - Enable pull-up

#define DDRB_INIT  0x0E
#define PORTB_INIT 0xF3

#define DIR_RIGHT PORTD, PD0	// OUT - Low
#define PWM_RIGHT PORTD, PD1	// OUT - low
#define DIR_LEFT  PORTD, PD2	// OUT - low
#define PWM_LEFT  PORTD, PD3	// OUT - low
#define DIS_LEFT  PORTD, PD4	// OUT - high
#define RC_AUX    PIND,  PD5	// IN  - Enable Pull-up
#define RC_LEFT   PIND,  PD6	// IN  - Enable Pull-up

#define DDRD_INIT  0x1F
#define PORTD_INIT 0x70
;
; Register Definitions
;
#define NotUsed0        R0
#define zero            R1

#define left_pwm        R2      // Working PWM values
#define right_pwm       R3
#define tmpw_lo		R4
#define tmpw_hi		R5
#define sreg_save	R6
#define left_hold       R7      // Holding registers for last computed A & B Pulse width
#define right_hold      R8
#define NotUsed09         R9
#define NotUsed10       R10
#define NotUsed11       R11
#define NotUsed12       R12
#define NotUsed13       R13
#define NotUsed14       R14
#define NotUsed15       R15

#define int_tmp		R16     // Interrupt handler temp.
#define p1              R17
#define aux_hold	R18
#define count           R19     // PWM Counter
#define flags           R20
#define     LEFT_ST     0   	// Bit Values
#define     LEFT_END    1
#define     RIGHT_ST    2
#define     RIGHT_END   3
#define     AUX_ST      4
#define     AUX_END     5

#define Valid		R21
#define rc_aux_lo       R22	// Counters for R/C pulse width measurement
#define rc_aux_hi       R23
#define rc_right_lo     R24
#define rc_right_hi     R25
#define rc_left_lo      R26
#define rc_left_hi      R27
#define sw_tmp          R28
#define sw_tmp_lo	R28
#define sw_tmp_hi	R29

#define left_pwm_hold	R30	// Holding registers while calculating PWM values
#define right_pwm_hold	R31
;+---------------------------------------------------------------------
;
; Init section of code with reset and interrupt vectors.
;
        .section        .init
        .func   _init
_init:
        rjmp    reset
        rjmp    halt
#if defined(at90s1200)
        .org    0x4
        rjmp    _overflow0_
#elif defined(at90s2313)
        .org    0xC
        rjmp    _overflow0_
#else
# error MCU != 1200 or 2313
#endif
halt:
        cli
        rjmp    .

        .endfunc
;-----------------------------------------------------------------------

	.section        .text
;
; main routine
;
        .func   reset
        .global reset
reset:
	cli			; Reset reset hardware in case we
	clr	R31		; didn't get here via a hardware
	out	WDTCR, R31	; reset
        ;
	; Initialize all working registers to zero
	;
	ldi     R30, 29         ; Point to R29
lpz:    st      Z, R31
        dec     R30             ; Zap all registers
        brne    lpz
	;
#if defined(at90s2313)
        out     0x3C, R16       ; Initialize Stack
        ldi     R16, 0xDF
        out     0x3D, R16
#endif
        ;
        ; Initialize all CPU hardware peripherals
        ;
	ldi	R16, PORTD_INIT	; set output state before direction.
	out	PORTD, R16	; at reset output is low, direction is in.
	ldi	R16, DDRD_INIT
	out	DDRD, R16
	ldi	R16, PORTB_INIT	; ditto
	out	PORTB, R16
	ldi	R16, DDRB_INIT
	out	DDRB, R16

	ldi	R16, TIMSK_INIT	; yeah, this is just a bit set...
	out	TIMSK, R16
	ldi	R16, TCNT0_INIT
	out	TCNT0, R16
	ldi	R16, TCCR0_INIT
	out	TCCR0, R16

	ldi	R16, MCUCR_INIT	; set up sleep mode & Extn interrupt
	out	MCUCR, R16

        sei

#if defined(OSMC)
	;
	; Initialize the OSMC by delaying 1/2 second, then toggling
	; the disable lines.
	;
        ldi	sw_tmp_lo, lo8(1500)	; ~3khz PWM rate
        ldi	sw_tmp_hi, hi8(1500)
StartDelay:
	sleep
	tst	count
	brne	StartDelay
	sbiw	sw_tmp_lo, 1
	brne	StartDelay
	cbi	DIS_LEFT
	cbi	DIS_RIGHT
	sbi	DIS_LEFT
	sbi	DIS_RIGHT
#endif
	;
	; Clear R/C flags and reset counters
	;
	clr	flags
        ldi     rc_left_lo,  lo8(RC_OFFSET)
	ldi	rc_left_hi,  hi8(RC_OFFSET)
        ldi     rc_right_lo, lo8(RC_OFFSET)
	ldi	rc_right_hi, hi8(RC_OFFSET)
        ldi     rc_aux_lo,   lo8(RC_OFFSET)
	ldi	rc_aux_hi,   hi8(RC_OFFSET)
	;
	; Initialize holding registers to zero input in case of missing cable.
	;
;	ldi	aux_hold, RC_RANGE/2    	; lb 20020412
;	mov	left_hold, aux_hold
;	mov	right_hold, aux_hold
	;
	; Enable Watchdog
	;
	ldi	R16, WDTCR_INIT
	wdr			; Reset internal counter
	out	WDTCR, R16

forever:
	sleep			; Wait for an interrupt
;	tst     count
;	brne    forever
	;
	; Analyse the RC pulse input and set PWM output.
	;
	mov	sw_tmp, flags
	cbr	sw_tmp, (BV(LEFT_END) | BV(RIGHT_END) | BV(AUX_END))
	breq	forever			; skip if nothing to do

	ldi	R30, lo8(RC_RANGE+1)
	ldi	R31, hi8(RC_RANGE+1)

sw_check_left_pulse:
        sbrs    flags, LEFT_END
        rjmp    sw_check_right_pulse

	cp	rc_left_lo, R30
	cpc	rc_left_hi, R31
	brlo	1f
	clr	Valid			; bogus pulse
	rjmp	2f
1:
	inc	Valid
	cpi	Valid, VALID_PULSE
	brlo	2f
	ldi	Valid, VALID_PULSE
	mov	p1, rc_left_lo		; 0-200 (RC_RANGE)
	rcall	deadband		; convert to signed value & apply deadband
        mov     left_hold, p1
	wdr				; Reset watchdog with each valid pulse
2:
        ldi     rc_left_lo, lo8(RC_OFFSET)
	ldi	rc_left_hi, hi8(RC_OFFSET)
        cbr     flags, BV(LEFT_ST) | BV(LEFT_END)	; lb 20020412

sw_check_right_pulse:
        sbrs    flags, RIGHT_END
        rjmp    sw_check_aux_pulse

	cp	rc_right_lo, R30
	cpc	rc_right_hi, R31
	brlo	1f
	clr	Valid			; bogus pulse
	rjmp	2f
1:
	inc	Valid
	cpi	Valid, VALID_PULSE
	brlo	2f
	ldi	Valid, VALID_PULSE
	mov	p1, rc_right_lo
	rcall	deadband
        mov     right_hold, p1
	wdr				; Reset watchdog with each valid pulse
2:
        ldi     rc_right_lo, lo8(RC_OFFSET)
	ldi	rc_right_hi, hi8(RC_OFFSET)
        cbr     flags, BV(RIGHT_ST) | BV(RIGHT_END)

sw_check_aux_pulse:
        sbrs    flags, AUX_END
        rjmp    SetPwmValues

	cp	rc_aux_lo, R30
	cpc	rc_aux_hi, R31
	brlo	1f
	clr	Valid			; bogus pulse
	rjmp	2f
1:
	inc	Valid
	cpi	Valid, VALID_PULSE
	brlo	2f
	ldi	Valid, VALID_PULSE
        mov     aux_hold, rc_aux_lo
	wdr				; Reset watchdog with each valid pulse
2:
        ldi     rc_aux_lo, lo8(RC_OFFSET)
	ldi	rc_aux_hi, hi8(RC_OFFSET)
        cbr     flags, BV(AUX_ST) | BV(AUX_END)

SetPwmValues:
        sbic    MIXED                   ; If jumpered, then mix
        rjmp    straight

        mov	tmpw_lo, right_hold	; Sign extend steering
        clr	tmpw_hi
        sbrc	tmpw_lo, 7
        com	tmpw_hi

	mov	sw_tmp_lo, left_hold		; Sign extend throttle
	rcall	Sign_Extend_sw_tmp
	add	sw_tmp_lo, tmpw_lo      	; compute mixed signal
	adc	sw_tmp_hi, tmpw_hi
        rcall   limit           	; Limit to +/- 1/2 the range.
        mov     left_pwm_hold, sw_tmp_lo

	mov	sw_tmp_lo, left_hold		; Sign extend throttle
	rcall	Sign_Extend_sw_tmp
	sub	sw_tmp_lo, tmpw_lo      	; compute mixed signal
	sbc	sw_tmp_hi, tmpw_hi
        rcall   limit
        mov     right_pwm_hold, sw_tmp_lo
        rjmp    sw_set_drive_bits

straight:
	mov	sw_tmp_lo, left_hold
	rcall	Sign_Extend_sw_tmp
	rcall	limit
	mov	left_pwm_hold, sw_tmp_lo

	mov	sw_tmp_lo, right_hold
	rcall	Sign_Extend_sw_tmp
	rcall	limit
	mov	right_pwm_hold, sw_tmp_lo

sw_set_drive_bits:
	cpi	Valid, VALID_PULSE
	brsh	sw_check_flip
	clr	left_pwm_hold		; If not valid, kill output
	clr	right_pwm_hold
	ldi	aux_hold, RC_RANGE/2	; Set for both off
sw_check_flip:
	sbic	FLIP			; FLIP swaps left and right drive
	rjmp	1f			; and reversed direction.
	mov	R0, left_pwm_hold
	mov	left_pwm_hold, right_pwm_hold
	mov	right_pwm_hold, R0
	neg	right_pwm_hold		; 20020512 lb
	neg	left_pwm_hold
1:	;
	; Set left side bits
	;
        tst     left_pwm_hold
        brlt    1f
	cbi     DIR_LEFT		; Set forward
	brne	2f			; if drive == 0
#if defined(OSMC)
	sbic	BRAKE			;   if brake == 1
	sbi	DIS_LEFT		;   coast (disable)
#elif defined(LMD18200)
	sbi	DIS_LEFT		;
	sbis	BRAKE
	ldi	left_pwm_hold, PWM_RANGE
#else
# error No H-Bridge Defined
#endif
        rjmp    3f
1:
#if defined(OSMC)
	subi	left_pwm_hold, lo8(-PWM_RANGE)		; invert and reverse PWM
#elif defined(LMD18200)
	neg	left_pwm_hold		; lb 2/22/02 Make value posative
#else
# error No H-Bride Defined
#endif
	sbi     DIR_LEFT		; set DIR bit for reverse
2:
	cbi	DIS_LEFT		; Enable output
3:
	mov	left_pwm, left_pwm_hold	; transfer PWM value
	;
	; Set Right side bits.
	;
        tst     right_pwm_hold
        brlt    1f
        cbi     DIR_RIGHT
	brne	2f
#if defined(OSMC)
	sbic	BRAKE			;   if brake == 1
	sbi	DIS_RIGHT		;   coast (disable)
#elif defined(LMD18200)
	sbi	DIS_RIGHT		;
	sbis	BRAKE
	ldi	right_pwm_hold, PWM_RANGE
#else
# error No H-Bridge Defined
#endif
        rjmp    3f
1:
#if defined(OSMC)
	subi	right_pwm_hold, lo8(-PWM_RANGE)		; OSMC needs PWM inverted
#elif defined(LMD18200)
	neg	right_pwm_hold		; Intelligent H-Bridges don't. lb 2/22/02
#else
# error No H-Bridge Defined
#endif
	sbi     DIR_RIGHT
2:
	cbi	DIS_RIGHT
3:
	mov	right_pwm, right_pwm_hold
	;
	; Set the High or Low AUX output based upon
	; stick postion.
	;
	cpi	aux_hold, AUX_LOW_LIM
	brsh	1f
	cbi	AUX_HI			; SN75477 - inverted logic
	sbi	AUX_LO
	rjmp	set_drive_done
1:
	cpi	aux_hold, AUX_HIGH_LIM
	brlo	1f
	cbi	AUX_LO			; SN75477 - inverted logic
	sbi	AUX_HI
	rjmp	set_drive_done
1:
	cbi	AUX_HI			; Disable both outputs.
	cbi	AUX_LO
set_drive_done:
	rjmp	forever

        .endfunc			; .func reset

;+----------------------------------------------------------
;
; int Sign_Extend_R30
;
; Sign extends R30 into R31
;
;-----------------------------------------------------------

	.func	Sign_Extend_sw_tmp
Sign_Extend_sw_tmp:
	clr	sw_tmp_hi
	sbrc	sw_tmp_lo, 7
	com	sw_tmp_hi
	ret
	.endfunc

;+----------------------------------------------------------
;
; char limit(char)
;
; Limits sw_tmp_hi:sw_tmp_lo to +/- PWM_RANGE
;
;-----------------------------------------------------------

	.func	limit
limit:
	;
	; Now clip range to +/- PWM count
	;
	cpi	sw_tmp_lo, lo8(PWM_RANGE)
	ldi	p1, hi8(PWM_RANGE)
	cpc	sw_tmp_hi, p1
;	sbci	sw_tmp_hi, hi8(PWM_RANGE)
	brlt	1f
	ldi	sw_tmp_lo, PWM_RANGE
1:
	cpi	sw_tmp_lo, lo8(-PWM_RANGE)
	ldi	p1, hi8(-PWM_RANGE)
	cpc	sw_tmp_hi, p1
;	sbci	sw_tmp_hi, hi8(-PWM_RANGE)
	brge	1f
	ldi	sw_tmp_lo, lo8(-PWM_RANGE)
1:
	ret
	.endfunc			; .func limit
;+------------------------------------------------------------
;
; char deadband(char)
;
; Apply deadband adjustment to input signal.  Deadband shifts
; signal so output is a linear function of +/-0 to max
;
;-------------------------------------------------------------
	.func	deadband
deadband:
	subi	p1, RC_RANGE/2	; convert to signed value
	;
	; Set deadband based upon user jumper select
	;
	ldi	sw_tmp, DEADBAND_HIGH
	sbic	DEADBAND
	ldi	sw_tmp, DEADBAND_LOW
	;
	; Deadband adjustment also shifts output (expands range)
	;
	cp	p1, sw_tmp	; if (p1 >= DB)
	brge	1f		;   p1 -= DB
	neg	sw_tmp		; else if (p1 < -DB)
	cp	p1, sw_tmp	;   p1 -= -DB
	brlt	1f		; else
	mov	p1, sw_tmp	;   p1 = 0
1:
	sub	p1, sw_tmp
	ret
	.endfunc


;+---------------------------------------------------------
;
; Timer Interrupt handler
;
; Reload timer, examine inputs, drive outputs, etc.
; 44-46 cycles/60  or ~25% available @ 12mhz
;
; I.e. this puppy executes, on the average, 15 instructions
; in the main loop between interrupts!
;
;----------------------------------------------------------
	.func   _overflow0_
_overflow0_:
	in      sreg_save, SREG
;	ldi     int_tmp, TCNT0_INIT	; 9 cycles
	in	int_tmp, TCNT0		; 10 cycles
	subi	int_tmp, (CPUCLK/TICKRATE)-3
	out	TCNT0, int_tmp
;
; pwm_control
;
; Counts between 0 and MAXCOUNT
; compares pwm values with count, setting/clearing
; the PWM bit as needed. Not glitch free.
;
; 12-14 cycles
;
	inc     count
	cpi	count, PWM_RANGE
	brne	1f
	clr	count			; 4
1:
        cp      count, left_pwm
        brge    1f
        sbi     PWM_LEFT
        rjmp    2f
1:      cbi     PWM_LEFT		; 4/5
2:
        cp      count, right_pwm
        brge    1f
        sbi     PWM_RIGHT
        rjmp    2f
1:      cbi     PWM_RIGHT
2:
;
; Logic to measure the pulse width of the RC input
; called every 5us.  Relies on control softare to clear
; bits between cycles. 5 cycles inactive, 7 cycles active.
; only one channel active at a time (r/c radio signal)
;
; 17 cycles
;
rc_read:
        sbis    RC_LEFT
        rjmp    1f
	subi	rc_left_lo, lo8(-1)
	sbci	rc_left_hi, hi8(-1)
        sbr     flags, BV(LEFT_ST)
        rjmp    2f
1:
        sbrc    flags, LEFT_ST
        sbr     flags, BV(LEFT_END)
2:
        sbis    RC_RIGHT
        rjmp    1f
	subi	rc_right_lo, lo8(-1)
	sbci	rc_right_hi, hi8(-1)
        sbr     flags, BV(RIGHT_ST)
        rjmp    2f
1:
        sbrc    flags, RIGHT_ST
        sbr     flags, BV(RIGHT_END)
2:
        sbis    RC_AUX
        rjmp    1f
	subi	rc_aux_lo, lo8(-1)
	sbci	rc_aux_hi, hi8(-1)
        sbr     flags, BV(AUX_ST)
        rjmp    2f
1:
        sbrc    flags, AUX_ST
        sbr     flags, BV(AUX_END)
2:
        out     SREG, sreg_save		; 5 cycles to return
        reti
        .endfunc

